"""Some addtional utilities."""

from typing import Any, Union

import torch


def to_device(
    tensors: Any, device: Union[str, torch.device], non_tensors: str = "error"
) -> Any:
    """Transfer a whole container of tensors to a device.

    The function takes an arbitrarily nested structure of uples, lists, and dicts whose
    entries at some point are tensors (or arbitrary other data types, see argument
    'non_tensors') and returns the same nested structure but with all tensors
    transfered to the given device.

    Args:
        tensors (Any): The container of (eventually) tensors to be transfered.
        device (Union[str, torch.device]): The target device.
        non_tensors (str, optional): A string describing the behaviour of the function
          when a non-tensor is encountered. If 'error' raises a ValueError, if 'ignore'
          the value is returned as is, if 'drop' the value is not included in the output
          (Note that this may subtly change the nested structure of the output since
          lists and tuples may be shorter than in the input and dicts may be missing
          keys). Defaults to "error".

    Raises:
        ValueError: If 'non_tensors' is 'error' and a value not of type list, tuple,
          dict, or tensor is encountered.
        ValueError: If an unsupported option for 'non_tensor' is given.

    Returns:
        Any: Same type and nested structure as the input but with all tensors
        on the given device.
    """
    if isinstance(tensors, torch.Tensor):
        return tensors.to(device=device)
    elif isinstance(tensors, list):
        return [
            t_on_device
            for t in tensors
            if (t_on_device := to_device(t, device=device, non_tensors=non_tensors))
            is not None
        ]
    elif isinstance(tensors, tuple):
        return tuple(
            t_on_device
            for t in tensors
            if (t_on_device := to_device(t, device=device, non_tensors=non_tensors))
            is not None
        )
    elif isinstance(tensors, dict):
        return {
            key: t_on_device
            for key, t in tensors.items()
            if (t_on_device := to_device(t, device=device, non_tensors=non_tensors))
            is not None
        }
    else:
        if non_tensors == "error":
            raise ValueError(f"Encountered non-tensor type {type(tensors)}.")
        elif non_tensors == "ignore":
            return tensors
        elif non_tensors == "drop":
            return None
        else:
            raise ValueError(f"Unknown value for argument 'non_tensors' {non_tensors}")



"""The actual SWEM-Model."""

from itertools import chain
from typing import Optional, Tuple

import torch
from torch import nn

from swem.models.pooling import SwemPoolingLayer


class Swem(nn.Module):
    """Simple Word Embedding model (see
    `Baselines need more love <https://arxiv.org/abs/1808.09843>`_ ).

    The model consists of an embedding layer, a feed forward network that is applied
    separately to each word vector, a pooling layer that pools the vectors belonging to
    the same text into a single vector, and another feed forward network that is applied
    to this pooled vector.

    Args:
        embedding (nn.Embedding): The embedding layer used by the model.
        pooling_layer (nn.Module): The pooling layer to be used by the model.
        pre_pooling_dims (Optional[Tuple[int, ...]]): Intermediate dimensions for the
          feed forward network applied to the input.
        post_pooling_dims (Optional[Tuple[int, ...]]): Intermediate dimensions for the
          feed forward network applied to the output of the pooling layer.
        dropout (float): Dropout probability after each layer in both feed forward
          subnetworks.

    Shapes:
        - input: :math:`(\\text{batch_size}, \\text{seq_len})`
        - output: :math:`(\\text{batch_size}, \\text{enc_dim})`, where
          :math:`\\text{enc_dim}` is the last of the post_pooling_dims (if given,
          otherwise the last of the pre_pooling_dims or failing that the
          embedding_dimension).
    """

    def __init__(
        self,
        embedding: nn.Embedding,
        pooling_layer: SwemPoolingLayer,
        pre_pooling_dims: Optional[Tuple[int, ...]] = None,
        post_pooling_dims: Optional[Tuple[int, ...]] = None,
        dropout: float = 0.2,
    ):
        super().__init__()
        self.embedding = embedding
        self.pooling_layer = pooling_layer
        if pre_pooling_dims is None:
            self.pre_pooling_trafo: nn.Module = nn.Identity()
        else:
            pre_dims = [embedding.embedding_dim, *pre_pooling_dims]
            self.pre_pooling_trafo = nn.Sequential(
                *chain(
                    *[
                        (nn.Linear(dim_in, dim_out), nn.ReLU(), nn.Dropout(dropout))
                        for dim_in, dim_out in zip(pre_dims[:-1], pre_dims[1:])
                    ]
                )
            )

        if post_pooling_dims is None:
            self.post_pooling_trafo: nn.Module = nn.Identity()
        else:
            pooling_dim = (
                embedding.embedding_dim
                if pre_pooling_dims is None
                else pre_pooling_dims[-1]
            )
            if len(post_pooling_dims) == 1:
                self.post_pooling_trafo = nn.Linear(pooling_dim, post_pooling_dims[0])
            else:
                post_dims = [pooling_dim, *post_pooling_dims[:-1]]
                self.post_pooling_trafo = nn.Sequential(
                    *chain(
                        *[
                            (nn.Linear(dim_in, dim_out), nn.ReLU(), nn.Dropout(dropout))
                            for dim_in, dim_out in zip(post_dims[:-1], post_dims[1:])
                        ]
                    ),
                    torch.nn.Linear(post_dims[-1], post_pooling_dims[-1]),
                )

    def forward(self, input: torch.Tensor) -> torch.FloatTensor:
        output = self.embedding(input)
        output = self.pre_pooling_trafo(output)
        output = self.pooling_layer(output)
        output = self.post_pooling_trafo(output)
        return output
